package dev.fitko.fitconnect.client;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataEncryptionIssue;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataSchemaViolation;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.AttachmentCreationException;
import dev.fitko.fitconnect.api.exceptions.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.EventLogException;
import dev.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.SubmissionNotCreatedException;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.Attachment;
import dev.fitko.fitconnect.client.sender.model.EncryptedAttachment;
import dev.fitko.fitconnect.client.sender.model.SendableEncryptedSubmission;
import dev.fitko.fitconnect.client.sender.model.SendableSubmission;
import dev.fitko.fitconnect.client.testutil.LogCaptor;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.nio.file.Path;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.core.IsEqual.equalTo;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

public class SenderClientTest {

    private final Sender senderMock;
    private final SenderClient senderClient;

    private static final LogCaptor logs = new LogCaptor();

    private final UUID destinationId = UUID.randomUUID();

    private final RSAKey publicKey = generateRsaKey(2048).toPublicJWK();

    @BeforeEach
    public void clearLogs(){
        logs.clearEvents();
    }

    public SenderClientTest() throws JOSEException {
        senderMock = mock(Sender.class);
        senderClient = new SenderClient(senderMock);
        setupTestMocks(destinationId);
    }

    @Test
    void testSendSubmissionWithJsonData() {

        // Given
        final RSAKey publicKey = this.publicKey;

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(announcedSubmission.getSubmissionId());

        final SubmissionSchema jsonSchema = new SubmissionSchema();
        jsonSchema.setSchemaUri(URI.create("https://json.mimetype"));
        jsonSchema.setMimeType(MimeType.APPLICATION_JSON);
        final Destination destination = getDestination(destinationId, jsonSchema);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.sendSubmission(any())).thenReturn(expectedSubmission);
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);
        when(senderMock.validateJsonFormat(any())).thenReturn(ValidationResult.ok());
        when(senderMock.validateMetadata(any())).thenReturn(ValidationResult.ok());


        // When
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://json.mimetype"))
                .addAttachment(Attachment.fromPath(Path.of("src/test/resources/attachment.txt"), "text/plain"))
                .build();

        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNotNull(sentSubmission);
        assertThat(sentSubmission.getSubmissionId(), equalTo(expectedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    void testSendSubmissionWithXmlData() {

        // Given
        final RSAKey publicKey = this.publicKey;

        final var destinationId = UUID.randomUUID();
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(announcedSubmission.getSubmissionId());

        final SubmissionSchema xmlSchema = new SubmissionSchema();
        xmlSchema.setSchemaUri(URI.create("https://xml.mimetype"));
        xmlSchema.setMimeType(MimeType.APPLICATION_XML);
        final Destination destination = getDestination(destinationId, xmlSchema);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.sendSubmission(any())).thenReturn(expectedSubmission);
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);
        when(senderMock.validateMetadata(any())).thenReturn(ValidationResult.ok());
        when(senderMock.validateXmlFormat(any())).thenReturn(ValidationResult.ok());

        // When
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setXmlData("<xml><test>data</test></xml>", URI.create("https://xml.mimetype"))
                .addAttachment(Attachment.fromPath(Path.of("src/test/resources/attachment.txt"), "text/plain"))
                .build();

        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNotNull(sentSubmission);
        assertThat(sentSubmission.getSubmissionId(), equalTo(expectedSubmission.getSubmissionId()));
        assertThat(destinationId, equalTo(announcedSubmission.getDestinationId()));
    }

    @Test
    void testMimeTypeDetection() throws Exception {

        // Given
        final UUID destinationId = setupTestMocks();
        final File file1 = new File("src/test/resources/text_attachment_with_no_filetype");
        final File file2 = new File("src/test/resources/pdf_attachment_with_no_extension");

        // When
        final var submission = SubmissionBuilder.Builder()
                .withAttachments(List.of(file1, file2))
                .withJsonData("{}", URI.create("https://mimetype.test.de"))
                .withDestination(destinationId)
                .withServiceType("name", "urn:de:fim:leika:leistung:99400048079000")
                .build();

        final var sentSubmission = senderClient.submit(submission);

        // Then
        assertNotNull(sentSubmission);
        logs.assertContains("Detected attachment mime-type text/plain");
        logs.assertContains("Detected attachment mime-type application/pdf");
    }

    @Test
    void testMissingDataMimeTypeSchema() {

        // Given
        final var destinationId = UUID.randomUUID();

        final var destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

        final var destination = new Destination();
        destination.setServices(Set.of(destinationService));
        destination.setDestinationId(destinationId);

        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = this.publicKey;

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.validateJsonFormat(any())).thenReturn(ValidationResult.ok());
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);

        // When
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"))
                .build();

        final var sentSubmission = senderClient.send(submission);

        // Then
        assertThat(exception.getMessage(), equalTo("Combination of provided MIME type 'application/json' and schema URI " +
                "'https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json' is not allowed by the destination"));
    }

    @Test
    void testWithEmptyAttachments() {

        // Given
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .addAttachments(List.of())
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNotNull(sentSubmission);
        logs.assertContains("No attachments to upload");
    }

    @Test
    void testWithAttachmentFileNull() {

        // Given
        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .addAttachment(null)
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNotNull(sentSubmission);
        logs.assertContains("No attachments to upload");
    }

    @Test
    void testWithAttachmentPathNotPresent() {

        // Given
        final var path = Path.of("/non/existing/path");

        // When
        final AttachmentCreationException exception = assertThrows(AttachmentCreationException.class, () -> Attachment.fromPath(path, "plain/text"));

        // Then
        assertThat(exception.getMessage(), containsString("Reading attachment from path '" + path + "' failed"));
    }

    @Test
    void testWithAttachmentInputStreamNotReadable() {

        // Given
        final var inputStream = new InputStream() {
            @Override
            public int read() throws IOException {
                throw new IOException("reading from stream failed");
            }
        };

        // When
        final AttachmentCreationException exception = assertThrows(AttachmentCreationException.class, () -> Attachment.fromInputStream(inputStream, "plain/text", "non-existing-file", "test"));

        // Then
        assertThat(exception.getMessage(), containsString("Attachment could not be read from input-stream"));
    }

    @Test
    void testEncryptionError() {

        // Given
        when(senderMock.encryptBytes(any(), any())).thenThrow(new EncryptionException("Encryption failed"));

        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNull(sentSubmission);
        logs.assertContains("Encrypting submission failed");
    }

    @Test
    void testAnnouncingSubmissionFailed() {

        // Given
        when(senderMock.createSubmission(any())).thenThrow(new SubmissionNotCreatedException("Announcing submission failed"));

        final var submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        // Then
        assertNull(sentSubmission);
        logs.assertContains("Failed to announce new submission");

    }

    @Test
    void testRestApiCallFailed() {

        // Given
        when(senderMock.createSubmission(any())).thenThrow(new RestApiException("Announcing submission failed"));

        final SendableSubmission submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        assertNull(sentSubmission);
        logs.assertContains("Sending submission failed");

    }

    @Test
    void testGettingEncryptionKeyFailed() {

        // Given

        when(senderMock.getEncryptionKeyForDestination(any())).thenThrow(new KeyNotRetrievedException("Getting encryption key failed"));

        // When
        final SendableSubmission submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        final var sentSubmission = senderClient.send(submission);

        assertNull(sentSubmission);
        logs.assertContains("Getting encryption key for destination " + destinationId + " failed");
    }

    @Test
    void testSendSubmissionWithEncryptedData() {

        // Given
        final SendableEncryptedSubmission submission = SendableEncryptedSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setEncryptedMetadata("$ncrypt$d m$t@d@t@")
                .setEncryptedData("$ncrypt$d d@t@")
                .addEncryptedAttachment(new EncryptedAttachment(UUID.randomUUID(), "@tt@chm$nt"))
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        assertNotNull(sentSubmission);
        logs.assertContains("SUCCESSFULLY HANDED IN SUBMISSION");
    }

    @Test
    void testPublicEncryptionKeyRetrieval() {

        // Given
        final var destinationId = UUID.randomUUID();
        final var destination = getDestination(destinationId);

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);

        // When
        final String encryptionKey = senderClient.getPublicKeyForDestination(destinationId);

        assertNotNull(encryptionKey);
        assertThat(encryptionKey, equalTo(publicKey.toJSONString()));
    }

    @Test
    void testFailOnInvalidMetadata() {

        // Given
        when(senderMock.validateMetadata(any())).thenReturn(ValidationResult.error(new ValidationException("invalid metadata")));

        final SendableSubmission submission = SendableSubmission.Builder()
                .setDestination(destinationId)
                .setServiceType("urn:de:fim:leika:leistung:99400048079000", "TestService")
                .setJsonData("{}", URI.create("https://mimetype.test.de"))
                .build();

        // When
        final var sentSubmission = senderClient.send(submission);

        assertNull(sentSubmission);
        logs.assertContains("Metadata does not match schema");
    }

    @Test
    void testGetValidEventLog() {

        // Given
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();
        final List<EventLogEntry> expectedLog = List.of(EventLogEntry.builder().build());
        when(senderMock.getEventLog(any(), any())).thenReturn(expectedLog);

        // When
        final List<EventLogEntry> eventLog = senderClient.getEventLog(caseId, destinationId);

        // Then
        assertThat(eventLog, is(expectedLog));

    }

    @Test
    void testGetInvalidEventLog() {

        // Given
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();
        when(senderMock.getEventLog(any(), any())).thenThrow(new EventLogException("Invalid log"));

        // When
        final EventLogException exception = assertThrows(EventLogException.class, () -> senderClient.getEventLog(caseId, destinationId));

        // Then
        assertThat(exception.getMessage(), containsString("Invalid log"));

    }

    @Test
    void testGetSubmissionStatus() {

        // Given
        final SentSubmission sentSubmission = SentSubmission.builder()
                .caseId(UUID.randomUUID())
                .submissionId(UUID.randomUUID())
                .destinationId(UUID.randomUUID())
                .authenticationTags(new AuthenticationTags())
                .build();

        final SubmissionStatus submissionStatus = new SubmissionStatus(Event.ACCEPT.getState(), Collections.emptyList());

        when(senderMock.getLastedEvent(any(), any(), any(), any())).thenReturn(submissionStatus);

        // When
        final SubmissionStatus statusForSubmission = senderClient.getStatusForSubmission(sentSubmission);

        // Then
        assertThat(statusForSubmission.getStatus(), is(Event.ACCEPT.getState()));

    }

    @Test
    void testGetSubmissionStatusWithProblems() {

        // Given
        final SentSubmission sentSubmission = SentSubmission.builder()
                .caseId(UUID.randomUUID())
                .submissionId(UUID.randomUUID())
                .destinationId(UUID.randomUUID())
                .authenticationTags(new AuthenticationTags())
                .build();

        final SubmissionStatus submissionStatus = new SubmissionStatus(Event.REJECT.getState(), List.of(new DataEncryptionIssue(), new DataSchemaViolation()));

        when(senderMock.getLastedEvent(any(), any(), any(), any())).thenReturn(submissionStatus);

        // When
        final SubmissionStatus statusForSubmission = senderClient.getStatusForSubmission(sentSubmission);

        // Then
        assertThat(statusForSubmission.getStatus(), is(Event.REJECT.getState()));
        assertThat(statusForSubmission.getProblems(), is(List.of(new DataEncryptionIssue(), new DataSchemaViolation())));

    }

    @Test
    void validateCallback() {

        when(senderMock.validateCallback(anyString(), anyLong(), anyString(), anyString())).thenReturn(ValidationResult.ok());

        final ValidationResult validationResult = senderClient.validateCallback("hmac", 0L, "body", "secret");

        verify(senderMock, times(1)).validateCallback(anyString(), anyLong(), anyString(), anyString());
        assertTrue(validationResult.isValid());
        assertFalse(validationResult.hasError());
    }

    private void setupTestMocks(final UUID destinationId) {

        final var destination = getDestination(destinationId);
        final var announcedSubmission = getAnnouncedSubmission(destinationId);

        final RSAKey publicKey = this.publicKey;

        when(senderMock.getDestination(any())).thenReturn(destination);
        when(senderMock.createSubmission(any())).thenReturn(announcedSubmission);
        when(senderMock.getEncryptionKeyForDestination(any())).thenReturn(publicKey);
        when(senderMock.validateMetadata(any())).thenReturn(ValidationResult.ok());
        when(senderMock.validateXmlFormat(any())).thenReturn(ValidationResult.ok());
        when(senderMock.validateJsonFormat(any())).thenReturn(ValidationResult.ok());
        when(senderMock.sendSubmission(any())).thenReturn(new Submission());

    }

    private SubmissionForPickup getAnnouncedSubmission(final UUID destinationId) {
        final SubmissionSchema submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("https://json-schema.fitco.de"));
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final DestinationService destinationService = new DestinationService();
        destinationService.setIdentifier("1");
        destinationService.setSubmissionSchemas(Set.of(submissionSchema));

        final var announcedSubmission = new SubmissionForPickup();
        announcedSubmission.setSubmissionId(UUID.randomUUID());
        announcedSubmission.setDestinationId(destinationId);
        return announcedSubmission;
    }

    private Destination getDestination(final UUID destinationId) {
        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://mimetype.test.de"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final var destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    private Destination getDestination(final UUID destinationId, final SubmissionSchema schema) {
        final var destinationService = new DestinationService();
        destinationService.setIdentifier("urn:de:fim:leika:leistung:99400048079000");
        destinationService.setSubmissionSchemas(Set.of(schema));

        final var destination = new Destination();
        destination.setDestinationId(destinationId);
        destination.setServices(Set.of(destinationService));
        return destination;
    }

    private static RSAKey generateRsaKey(final int size) throws JOSEException {
        return new RSAKeyGenerator(size)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();
    }
}